%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Implementation of readCurry of module AbsCurry (reading abstract Curry
% programs):

?- block 'prim_AbsCurry.readCurryFile'(-,?,?,?,?),
         'prim_AbsCurry.readCurryFile'(?,?,?,-,?).
'prim_AbsCurry.readCurryFile'(ProgString,_,'$io'(CurryProg),E0,E) :-
	cp_string(ProgString,ProgName),
	atom_chars(ProgNameAtom,ProgName),
	readACProg(ProgNameAtom,ACProg),
	!,
	abscy_transProg(ACProg,CurryProg),
	!, E0=E.

% read an abstract Curry program:
readACProg(Prog,ACProg) :-
	split2dirbase(Prog,ProgDir,ProgBase),
	loadPath(ProgDir,LP),
	readACProgInLoadPath(LP,ProgBase,ACProg).

readACProgInLoadPath([],ACFile,_) :-
	write('ERROR: AbstractCurry file '), write(ACFile),
	write(' not found!'), nl,
	!, fail.
readACProgInLoadPath([Dir|Dirs],ACFile,ACProg) :-
	(atom_chars(ACFile,[47|_])   %already absolute file name?
	 -> DirACFile=ACFile
	  ; appendAtom(Dir,'/',DirSlash),
	    appendAtom(DirSlash,ACFile,DirACFile)),
	(file_exists(DirACFile)
	 -> readACProgFile(DirACFile,ACProg)
	  ; readACProgInLoadPath(Dirs,ACFile,ACProg)).

readACProgFile(ACFile,ACProg) :-
	%write('>>> Reading '), write(ACFile), write(' ...'), nl,
	see(ACFile),
	read(ACProg),
	seen.


abscy_transProg('CProg'(ModName,Imports,Types,Funcs,Ops),
	      'AbsCurry.CProg'(FModName,FImports,CTypes,CFuncs,COps)) :-
	cp_string(FModName,ModName),
	map2(cp_string,FImports,Imports),
	map2(abscy_transTypeDecl,Types,CTypes),
	abscy_transFuncs(Funcs,CFuncs),
	map2(abscy_transOpDecl,Ops,COps).

abscy_transVisibility('CExported','AbsCurry.CExported').
abscy_transVisibility('CPrivate','AbsCurry.CPrivate').

abscy_transTypeDecl('CType'(Name,Visib,TParams,ConsDecls),
		  'AbsCurry.CType'(FName,CVisib,TParams,CConsDecls)) :-
	cp_string(FName,Name),
	abscy_transVisibility(Visib,CVisib),
	map2(abscy_transConsDecl,ConsDecls,CConsDecls).
abscy_transTypeDecl('CTypeSyn'(Name,Visib,TParams,TExp),
		  'AbsCurry.CTypeSyn'(FName,CVisib,TParams,CTExp)) :-
	cp_string(FName,Name),
	abscy_transVisibility(Visib,CVisib),
	abscy_transTypeExpr(TExp,CTExp).

abscy_transConsDecl('CCons'(Name,Arity,Visib,Types),
		  'AbsCurry.CCons'(FName,Arity,CVisib,CTypes)) :-
	cp_string(FName,Name),
	abscy_transVisibility(Visib,CVisib),
	map2(abscy_transTypeExpr,Types,CTypes).

abscy_transTypeExpr('CTVar'(N),'AbsCurry.CTVar'(N)).
abscy_transTypeExpr('CFuncType'(T1,T2),'AbsCurry.CFuncType'(CT1,CT2)) :-
	abscy_transTypeExpr(T1,CT1), abscy_transTypeExpr(T2,CT2).
abscy_transTypeExpr('CTCons'(Mod,Name,Ts),
		  'AbsCurry.CTCons'(FMod,FName,CTs)) :-
	cp_string(FMod,Mod),
	cp_string(FName,Name),
	map2(abscy_transTypeExpr,Ts,CTs).


abscy_transOpDecl('COp'(Name,Fixity,Prio),
		'AbsCurry.COp'(FName,CFixity,Prio)) :-
	cp_string(FName,Name),
	abscy_transFixity(Fixity,CFixity).

abscy_transFixity('CInfixOp','AbsCurry.CInfixOp').
abscy_transFixity('CInfixlOp','AbsCurry.CInfixlOp').
abscy_transFixity('CInfixrOp','AbsCurry.CInfixrOp').


abscy_transFuncs([],[]).
abscy_transFuncs([F|Fs],[TF|TFs]) :-
	abscy_transFuncDecl(F,TF),
	abscy_transFuncs(Fs,TFs).

abscy_transFuncDecl('CFunc'(Name,Arity,Visib,Type,Rule),
		  'AbsCurry.CFunc'(FName,Arity,CVisib,CType,CRule)) :-
	cp_string(FName,Name),
	abscy_transVisibility(Visib,CVisib),
	abscy_transTypeExpr(Type,CType),
	abscy_transRules(Rule,CRule).

abscy_transRules('CRules'(Eval,Rules),'AbsCurry.CRules'(CEval,CRules)) :-
	abscy_transEval(Eval,CEval),
	map2(abscy_transRule,Rules,CRules).
abscy_transRules('CExternal'(Name),'AbsCurry.CExternal'(FName)) :-
	cp_string(FName,Name).

abscy_transEval('CFlex','AbsCurry.CFlex').
abscy_transEval('CRigid','AbsCurry.CRigid').
abscy_transEval('CChoice','AbsCurry.CChoice').

abscy_transRule('CRule'(Pats,CondRhs,Locals),
	      'AbsCurry.CRule'(CPats,CCondRhs,CLocals)) :-
	map2(abscy_transPattern,Pats,CPats),
	map2(abscy_transCRhs,CondRhs,CCondRhs),
	map2(abscy_transLocalDecl,Locals,CLocals).

abscy_transLocalDecl('CLocalFunc'(Func),'AbsCurry.CLocalFunc'(CFunc)) :-
	abscy_transFuncDecl(Func,CFunc).
abscy_transLocalDecl('CLocalPat'(Type,Pat,Exp,Locals),
		   'AbsCurry.CLocalPat'(CType,CPat,CExp,CLocals)) :-
	abscy_transTypeExpr(Type,CType),
	abscy_transPattern(Pat,CPat),
	abscy_transExpr(Exp,CExp),
	map2(abscy_transLocalDecl,Locals,CLocals).
abscy_transLocalDecl('CLocalVar'(Type,Var),
		   'AbsCurry.CLocalVar'(CType,Var)) :-
	abscy_transTypeExpr(Type,CType).

abscy_transCRhs('(,)'(Exp1,Exp2),'(,)'(CExp1,CExp2)) :-
	abscy_transExpr(Exp1,CExp1), abscy_transExpr(Exp2,CExp2).


abscy_transExpr('CVar'(N),'AbsCurry.CVar'(N)) :- !.
abscy_transExpr('CLit'(L),'AbsCurry.CLit'(CL)) :- !, abscy_transLit(L,CL).
abscy_transExpr('CSymbol'(Mod,Name),'AbsCurry.CSymbol'(FMod,FName)) :-
	cp_string(FMod,Mod), cp_string(FName,Name).
abscy_transExpr('CApply'(E1,E2),'AbsCurry.CApply'(CE1,CE2)) :- !,
	abscy_transExpr(E1,CE1), abscy_transExpr(E2,CE2).
abscy_transExpr('CLambda'(Pats,E),'AbsCurry.CLambda'(CPats,CE)) :- !,
	map2(abscy_transPattern,Pats,CPats),
	abscy_transExpr(E,CE).
abscy_transExpr('CLetDecl'(Locals,E),'AbsCurry.CLetDecl'(CLocals,CE)) :- !,
	map2(abscy_transLocalDecl,Locals,CLocals),
	abscy_transExpr(E,CE).
abscy_transExpr('CListComp'(E,Stats),'AbsCurry.CListComp'(CE,CStats)) :- !,
	abscy_transExpr(E,CE),
	map2(abscy_transStat,Stats,CStats).
abscy_transExpr('CDoExpr'(Stats),'AbsCurry.CDoExpr'(CStats)) :- !,
	map2(abscy_transStat,Stats,CStats).
abscy_transExpr('CCase'(E,Cs),'AbsCurry.CCase'(CE,CCs)) :- !,
	abscy_transExpr(E,CE),
	map2(abscy_transBranch,Cs,CCs).
abscy_transExpr(E,E) :-
	write(user_error,'*** Internal error in AbstractCurry file: Illegal expression "'),
	write(user_error,E),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

abscy_transStat('CSExpr'(E),'AbsCurry.CSExpr'(CE)) :- abscy_transExpr(E,CE).
abscy_transStat('CSPat'(Pattern,E),'AbsCurry.CSPat'(CPattern,CE)) :-
	abscy_transPattern(Pattern,CPattern),
	abscy_transExpr(E,CE).
abscy_transStat('CSLet'(Locals),'AbsCurry.CSLet'(CLocals)) :-
	map2(abscy_transLocalDecl,Locals,CLocals).

abscy_transBranch('CBranch'(Pattern,E),
		'AbsCurry.CBranch'(CPattern,CE)) :-
	abscy_transPattern(Pattern,CPattern),
	abscy_transExpr(E,CE).

abscy_transPattern('CPVar'(N),'AbsCurry.CPVar'(N)) :- !.
abscy_transPattern('CPLit'(L),'AbsCurry.CPLit'(CL)) :- !, abscy_transLit(L,CL).
abscy_transPattern('CPComb'(Mod,Name,Pats),'AbsCurry.CPComb'(CMod,CName,CPats)) :-
	!,
	cp_string(CMod,Mod),
	cp_string(CName,Name),
	map2(abscy_transPattern,Pats,CPats).
abscy_transPattern(P,P) :-
	write(user_error,'*** Internal error in AbstractCurry file: Illegal pattern "'),
	write(user_error,P),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

abscy_transLit('CIntc'(I),'AbsCurry.CIntc'(I)) :- !.
abscy_transLit('CFloatc'(F),'AbsCurry.CFloatc'(F)) :- !.
abscy_transLit('CCharc'(N),'AbsCurry.CCharc'(C)) :- char_int(C,N), !.
abscy_transLit(Lit,Lit) :-
	write(user_error,'*** Internal error in AbstractCurry file: Illegal literal "'),
	write(user_error,Lit),
	write(user_error,'"!'), nl(user_error),
	pleaseReport, fail.

