%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Implementation of the port primitives.
%
% The clauses in this file are added to each compiled Curry program
% if the port communication primitives are used

:- use_module(library(system)).
:- use_module(library(sockets)).
:- use_module(prim_readshowterm). % for term en/decoding

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dynamic configurations for the port implementation:

:- dynamic only_localhost/1, tracePorts/1,
	   mySocket/1, incomingPortStreams/2.

% the the communication socket for this program
% (will be set by the first invocation of openPortOnSocket):
%mySocket('Ports.internalPort'(Host,SocketNr,MaxPortNumber,Socket)).

% lists of all pending incoming streams at our socket:
% IPNr is the port number of the streams
% Streams is a list of elements of the following form:
% message(Stream): a regular message to be read from Stream
% ping(Stream): a ping request for this port to be answered on Stream
%incomingPortStreams(IPNr,Streams).

% if machine is not connected to the inet (specified by PAKCS_LOCALHOST=yes)
% then local ports get the symbolic hostname "localhost"
only_localhost(LH) :-
	(environ('PAKCS_LOCALHOST',PLH) -> true ; PLH=no),
	(PLH=yes -> asserta(only_localhost(yes)), LH=yes
	          ; asserta(only_localhost(no)),  LH=no),
	retract((only_localhost(_) :- _,_)), !.


% trace all port communication if PAKCS_TRACEPORTS=yes
tracePorts(unknown).

ifTracePort(Goal) :-
	tracePorts(Status),
	(Status=unknown
	 -> retract(tracePorts(_)),
	    (environ('PAKCS_TRACEPORTS',PTP) -> true ; PTP=no),
	    (PTP=yes -> asserta(tracePorts(yes)), Goal
	              ; asserta(tracePorts(no)))
	  ; Status=yes -> Goal ; true).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definitions for port communication:
%
% internal representation of the handle of a port:
% 'Ports.internalPort'(Host,SNr,IPNr,Socket) where:
% - for external ports:
%   Host: Internet name of the host (a Curry list)
%   SNr: socket number of the port at Host (an integer)
%   IPNr: internal port number assigned to the port at Host (an integer)
%   Socket: Sicstus-Prolog handle of this socket
% - for internal ports:
%   Host: not used
%   SNr:  0
%   IPNr: not used
%   Socket: a list with an open end (the stream)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CURRENT RESTRICTIONS w.r.t. sending of logical variables on ports:
% 1. If a message sent over an external port contains lvars, then
%    the sender waits immediately after sending until these variables
%    are bound by the receiver
% 2. The receiver must bind the lvars of a received message
%    (incrementelly) to a ground term,
%    otherwise the binding will not be sent back


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ENCODING OF TERMS:
%
% Terms sent on an external port are encoded as strings as with
% the library function ReadShowTerm.showQTerm with the exception
% that logical variables are represented as (VAR i) where i is
% a unique index for the variable in this term.
%
% For instance, the message (M.Get "foo" x y x) is encoded as the
% string "(M.Get \"foo\" (VAR 0) (VAR 1) (VAR 0))".
%
% Sending bindings of logical variables:
%
% If a message contains logical variables, their bindings will be sent
% back through the same connection via the inverse stream, i.e.,
% if a message is received via an instream of a socket connection,
% the variable bindings are sent via the outstream of the same
% connection. Each binding is encoded as follows:
% "V<i>\n<b>EOT" where <i> is the unique index of the variable,
% <b> is the string encoding of the term to which variable <i> is bound,
% and EOT is the end-of-term marker (== char 0).
%
%
% SENDING MESSAGES:
% Messages are sent as strings using the encoding above to the
% socket assigned to the invocation of PAKCS. All port communication
% is done via this single socket. To distinguish different ports,
% each port has a unique number at the socket (this number is assigned
% by the primitive openPortOnSocket, see the implementation of
% prim_openPortOnSocket).
% The socket number is freely chosen or can be defined via the
% environment variable PAKCS_SOCKET.
% In order to determine the port to which a message
% is sent, the port number always precedes the message string.
% In particular, the structure of a message is as follows:
%
% N<pnr>\n<msg>EOT : A message encoded as string <msg> is sent to port <pnr>
%
% P<pnr>\n: Ping port <pnr>, i.e., tests whether there is a process which
%           is waiting for messages on this port. In this case, the line
%           "ok\n" is sent back on the same socket connection.
%
% EXAMPLE:
% As an example, we assume that we send the message (GetName "talk" v)
% (where GetName is defined in module "NameServer")
% to a server running at socket 1665 with port number 0. Then we open
% a connection to socket 1665 and send the following string:
% "N0\n(NameServer.GetName "talk" (VAR 0))\000" (remember: EOT == char 0)
% Since a logical variable (v) with index 0 has been sent, the sender
% listens on the input stream of this socket connection for the value
% of this variable. For instance, if the server binds this variable
% to the value 42, it sends the following string to the sender:
% "V0\n42\000". Since the sender has now received all variable values,
% he closes the connection.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%% open

?- block prim_openPort(?,?,?,-,?).
prim_openPort(RP,RS,R,E0,E) :- deref(RP,P), deref(RS,S),
	prim_openPortExec(P,S,R,E0,E).

prim_openPortExec(P,_,_,_,_) :-
	nonvar(P), !,
	write('ERROR: openPort: port is not a free variable'), nl,
	fail.
prim_openPortExec(_,S,_,_,_) :-
	nonvar(S), !,
	write('ERROR: openPort: stream is not a free variable'), nl,
	fail.
prim_openPortExec('Ports.internalPort'([],0,0,S),S,success,E0,E) :- E0=E.


?- block prim_openPortOnSocket(?,?,?,?,-,?).
prim_openPortOnSocket(RNewSocketNr,RNewPortNr,_,'$io'(Result),E0,E) :-
        deref(RNewSocketNr,NewSocketNr), deref(RNewPortNr,NewPortNr),
	(retract(mySocket('Ports.internalPort'(Host,SocketNr,MPN,Socket)))
	 -> MPN1 is MPN+1,
	    asserta(mySocket('Ports.internalPort'(Host,SocketNr,MPN1,Socket))),
	    ((NewSocketNr=SocketNr, NewPortNr=MPN1)
	     -> true
	      ; writeErr('ERROR: inconsistent socket/port numbers in openPortOnSocket!'),
		nlErr)
	  ; socket('AF_INET',Socket),
	    trySetSocketNumber(NewSocketNr),
	    % catch connection errors:
	    on_exception(Exc,socket_bind(Socket,'AF_INET'(AHost,NewSocketNr)),
                         (write('ERROR: openPortOnSocket "'),
			  write(NewSocketNr),
		          write('":'), nl, print_error(Exc), !, fail)),
   	    socket_listen(Socket,200),
	    atom2String(AHost,Host),
	    (var(NewPortNr) -> NewPortNr=0 ; true),
	    asserta(mySocket('Ports.internalPort'(Host,NewSocketNr,NewPortNr,
						  Socket)))
	),
	mySocket(Port),
	Port = 'Ports.internalPort'(CurrHost,CurrSocketNr,CurrPNr,_),
	ifTracePort((writeErr('TRACEPORTS: Listen on port '),
	             writeErr(CurrPNr),
	             writeErr(' at socket '),
	             writeErr(CurrSocketNr),
	             writeErr(' on host '),
		     (only_localhost(yes) -> writeErr('localhost')
	                                   ; string2Atom(CurrHost,ACurrHost),
			                     writeErr(ACurrHost)),
	             writeErr('...'), nlErr)),
	(compileWithSharing(function)
	 -> makeShare('Ports.basicServerLoop'(Port),Result)
	  ; Result = 'Ports.basicServerLoop'(Port)),
	E0=E.

% try setting socket number (if it is unbound) to the value of
% environment variable PAKCS_SOCKET, if it is defined:
trySetSocketNumber(SocketNr) :-
	var(SocketNr),
	environ('PAKCS_SOCKET',PSNr),
	atom_chars(PSNr,PSNrS), number_chars(SocketNr,PSNrS),
	!.
trySetSocketNumber(_).
	
% loop for reading the stream of incoming messages at a port:
?- block 'prim_basicServerLoop'(?,?,-,?).
'prim_basicServerLoop'(Port,Result,E0,E) :-
	readPort(Port,ReadResult,off),
	Result = [ReadResult|BSL],
	(compileWithSharing(function)
	 -> makeShare('Ports.basicServerLoop'(Port),BSL)
	  ; BSL = 'Ports.basicServerLoop'(Port)),
	E0=E.
 

% test whether a stream is instantiated in a particular time period:
?- block prim_timeoutOnStream(?,?,?,-,?).
prim_timeoutOnStream(RN,Stream,R,E0,E) :- deref(RN,N),
	prim_timeoutOnStream_exec(N,Stream,R,E0,E).

prim_timeoutOnStream_exec(_,Stream,'Nothing',E0,E) :- var(Stream), !, E0=E.
prim_timeoutOnStream_exec(Timeout,share(M),Result,E0,E) :- !,
        get_mutable(T,M),
	(T='$eval'(Exp) -> Result=Exp, E0=E
	 ; prim_timeoutOnStream_exec(Timeout,T,TOResult,E0,E),
	   (TOResult='Nothing'
	    -> Result=TOResult
	     ; (compileWithSharing(variable) -> propagateShare(TOResult,Result)
	                                      ; TOResult=Result),
	       update_mutable('$eval'(Result),M))).
prim_timeoutOnStream_exec(_,[],'Just'([]),E0,E) :- % stream already instantiated
        !, E0=E.
prim_timeoutOnStream_exec(_,[X|Xs],'Just'([X|Xs]),E0,E) :- !, E0=E.
prim_timeoutOnStream_exec(Timeout,'Ports.basicServerLoop'(Port),Result,E0,E) :-
	TOSec is Timeout//1000,
	TOMSec is (Timeout mod 1000) * 1000, % in microseconds
	(readPort(Port,ReadResult,TOSec:TOMSec)
	 -> (compileWithSharing(function)
	     -> makeShare('Ports.basicServerLoop'(Port),BSL)
	      ; BSL = 'Ports.basicServerLoop'(Port)),
	    Result='Just'([ReadResult|BSL])
	  ; Result='Nothing'),
	E0=E.
 

% primitives for binding an external logical variable:
?- block send_extvar_binding(?,-,?).
send_extvar_binding(Stream,A,Index,Done) :-
	evalToken(Eval),
	nf(A,NA,Eval,E),
        boolEq(NA,NA,_,E,E1), % wait for message to become bound
	eval_send_extvar_binding(Stream,NA,Index,E1,Done).
?- block eval_send_extvar_binding(?,?,?,-,?).
eval_send_extvar_binding(nostream,Term,Index,_,_) :-
	% this case should usually not occur since we send only ground
	% bindings for variables
	writeErr('ERROR: cannot send binding for variable '),
	writeErr(Index), writeErr(': '),
	writeErr(Term), nlErr,
	!.
eval_send_extvar_binding(Stream,Term,Index,_,Done) :-
	show4sending(Term,MsgString,_),
	ifTracePort((writeErr('TRACEPORTS: send binding for variable '),
                     writeErr(Index), writeErr(': '),
		     writeChars(user_error,MsgString), nlErr
		     %writeErr(Term), nlErr
		    )),
	number_chars(Index,IndexS),
	putChars(Stream,[86|IndexS]), % send variable index
	put(Stream,10),
	writeChars(Stream,MsgString),
	put(Stream,0), % send end-of-term marker
	flush_output(Stream),
	!, Done=done. % mark successful sending


readPort('Ports.internalPort'(Host,SNr,PNr,Socket),_,_):-
	var(Socket), !,
	string2Atom(Host,AHost),
	writeErr('ERROR: readPort: Port '),
	writeErr(SNr), writeErr('/'), writeErr(PNr),
	writeErr('@'), writeErr(AHost),
	writeErr(' only opened for writing!'), nlErr.
readPort('Ports.internalPort'(_,_,PNr,Socket),Msg,Timeout) :-
	% catch connection errors:
	on_exception(Exc,readFromSocket(PNr,Socket,Msg,Timeout),
                     (print_error(Exc)->Msg=[];Msg=[])), !.

% check the pending incoming streams on our socket for messages
% for a given port number and succeed with the stream; otherwise fail:
checkIncomingPortStreams(PNr,PortStream) :-
	retract(incomingPortStreams(PNr,[message(PortStream)|Streams])),
	asserta(incomingPortStreams(PNr,Streams)),
	!.
checkIncomingPortStreams(PNr,PortStream) :-
	retract(incomingPortStreams(PNr,[ping(PingStream)|Streams])),
	asserta(incomingPortStreams(PNr,Streams)),
	answerPingOnStream(PingStream),
	checkIncomingPortStreams(PNr,PortStream),
	!.

% answer a ping request on a stream:
answerPingOnStream(Stream) :-
	write(Stream,'ok'), put(Stream,10),
	flush_output(Stream),
	close(Stream),
	ifTracePort((writeErr('TRACEPORTS: Ping request answered.'),
		     nlErr)).

readFromSocket(PNr,Socket,Msg,Timeout) :-
	checkIncomingPortStreams(PNr,PortStream),
	!,
	readStreamUntilZero(PortStream,MsgString),
	(parse_received_message(PortStream,MsgString,Msg)
         -> ifTracePort((writeErr('TRACEPORTS: Received message: '),
                         putChars(user_error,MsgString), nlErr
		         %writeErr(Msg), nlErr
			))
          ; writeErr('ERROR: Illegal message received (ignored): '),
            putChars(user_error,MsgString), nlErr,
            readFromSocket(PNr,Socket,Msg,Timeout)),
	!.
readFromSocket(PNr,Socket,Msg,Timeout) :-
	%socket_select([Socket],PortStreams,Clients,1:0,[],_), % timeout: 1 sec
	%socket_select([Socket],PortStreams,Clients,off,[],_), % no timeout
	socket_select([Socket],PortStreams,Clients,Timeout,[],_),
	PortStreams=[PortStream], Clients=[Client],
	ifTracePort((writeErr('TRACEPORTS: Connection to client: '),
		     writeErr(Client), nlErr)),
	(readPortMessage(PNr,PortStream,MsgString)
	 -> (parse_received_message(PortStream,MsgString,Msg)
            -> ifTracePort((writeErr('TRACEPORTS: Received message: '),
                            putChars(user_error,MsgString), nlErr
		            %writeErr(Msg), nlErr
			   ))
             ; writeErr('ERROR: Illegal message received (ignored): '),
               putChars(user_error,MsgString), nlErr,
	       readFromSocket(PNr,Socket,Msg,Timeout))
	  ; % try reading next message:
	    readFromSocket(PNr,Socket,Msg,Timeout)),
	!.
readFromSocket(_,_Socket,_Msg,_) :-
	ifTracePort((writeErr('TRACEPORTS: Timeout!'),
		     nlErr)), fail.

% read from a (socket) stream a message with port number header;
% delay reading and fail if this message is not intended for the
% given port number:
readPortMessage(PNr,Stream,MsgString) :-
	readStreamLine(Stream,MsgHead),
	MsgHead = [FirstHeadChar|ReceivedPNrS],
	checkMessageHeader(FirstHeadChar,Stream,DecoratedStream),
	number_chars(ReceivedPNr,ReceivedPNrS),
	!,
	(ReceivedPNr=PNr
	 -> readPortMessageBody(DecoratedStream,MsgString)
	  ; % delay receiving of message and fail:
	    (retract(incomingPortStreams(ReceivedPNr,OldStreams))
	     -> append(OldStreams,[DecoratedStream],NewStreams)
	      ; NewStreams = [DecoratedStream] ),
	    asserta(incomingPortStreams(ReceivedPNr,NewStreams)),
	    !, fail).

checkMessageHeader(78,Stream,message(Stream)) :- !. % 78 = 'N'
checkMessageHeader(80,Stream,ping(Stream)) :- !.    % 80 = 'P'
checkMessageHeader(_,_,_) :-
	writeErr('ERROR: Illegal message header received.'),
	nlErr,
	!, fail.

readPortMessageBody(message(Stream),MsgString) :-
	readStreamUntilZero(Stream,MsgString).
readPortMessageBody(ping(Stream),_) :-
	answerPingOnStream(Stream),
	!, fail.
	
readStreamUntilZero(Str,String) :-
	get0(Str,C),
	(C==0 -> String=[]
	       ; readStreamUntilZero(Str,Rest),
	         String=[C|Rest]).



%%%% send

?- block prim_sendPort(?,?,?,-,?).
prim_sendPort(RMsg,RPort,R,E0,E) :-
	nf(RMsg,Msg,E0,E1), hnf(RPort,Port,E1,E2),
	prim_sendPortExec(Msg,Port,R,E2,E).

?- block prim_sendPortExec(-,?,?,?,?), prim_sendPortExec(?,-,?,?,?),
         prim_sendPortExec(?,?,?,-,?).
prim_sendPortExec(Msg,'Ports.internalPort'(_,0,_,Stream),success,E0,E) :-
	% send to internal port
	!,
	add2Stream(Stream,Msg), E0=E.
prim_sendPortExec('Ports.SP_Put'(Str),'Ports.internalPort'(_,-1,_,WIn),
		success,E0,E) :-
	% send to stream port
	!,
	boolEq(Str,Str,_,E0,E), % wait for string to become bound
	writeChars(WIn,Str), nl(WIn), flush_output(WIn),
	ifTracePort((writeErr('TRACEPORTS: SP_Put: '),
		     writeChars(user_error,Str),
		     nlErr)).
prim_sendPortExec('Ports.SP_GetLine'(Str),'Ports.internalPort'(WOut,-1,_,_),
		success,E0,E) :-
	% send to stream port
	!,
	readStreamLine(WOut,WOLine),
	map2(char_int,SPOutLine,WOLine),
	ifTracePort((writeErr('TRACEPORTS: SP_Get: '),
		     writeChars(user_error,SPOutLine),
		     nlErr)),
	constrEq(SPOutLine,Str,_,E0,E).	% unify SP_Get-Arg with read line
prim_sendPortExec('Ports.SP_GetChar'(Chr),'Ports.internalPort'(WOut,-1,_,_),
		success,E0,E) :-
	% send to stream port
	!,
	get0(WOut,NC), char_int(NC,C),
	ifTracePort((writeErr('TRACEPORTS: SP_GetChar: '),
		     writeErr(C),
		     nlErr)),
	constrEq(C,Chr,_,E0,E).	% unify SP_GetChar-Arg with read character
prim_sendPortExec('Ports.SP_EOF'(Bool),'Ports.internalPort'(WOut,-1,_,_),
		success,E0,E) :-
	% send to stream port
	!,
	(at_end_of_stream(WOut) -> EOF='True' ; EOF='False'),
	ifTracePort((writeErr('TRACEPORTS: SP_EOF: '),
		     writeErr(EOF),
		     nlErr)),
	constrEq(Bool,EOF,_,E0,E).	% unify SP_EOF-Arg with current value
prim_sendPortExec('Ports.SP_Close','Ports.internalPort'(WOut,-1,_,WIn),
		success,E0,E) :-
	% send to stream port
	!,
	close(WIn), close(WOut), % close input and output streams
	E0=E.
prim_sendPortExec(Msg,'Ports.internalPort'(_,-1,_,_),success,E0,E) :-
	% send to stream port
	!,
	writeErr('ERROR: wrong message received by stream port: '),
	writeErr(Msg), nlErr, E0=E.
prim_sendPortExec(Msg,'Ports.internalPort'(Host,SNr,PNr,_),success,E0,E) :-
	% send to external port
	% catch connection errors:
        string2Atom(Host,AHost),
	on_exception(Exc,send2Socket(AHost,SNr,PNr,Msg),
                     (write('ERROR: send to port '), write(PNr), write('@'),
		      write(AHost), write(':'), nl,
		      (print_error(Exc)->true;true))),	% send always succeeds!
	E0=E.

% append a message to the open end of a stream:
add2Stream(Str,Item) :- var(Str), !, Str = [Item|_].
add2Stream([],_) :-
	write('ERROR: send: stream has not a free variable at the end'),
	nl.
add2Stream([_|Str],Item) :- add2Stream(Str,Item).



%%%% ping

?- block prim_ping(?,?,?,?,-,?).
prim_ping(RN,RP,W,R,E0,E) :-
	deref(RN,N), deref(RP,P), prim_ping_exec(N,P,W,R,E0,E).

prim_ping_exec(_,'Ports.internalPort'(_,0,_,_),_,'$io'('Just'(0)),E0,E) :-
	% ping internal port
	!, E0=E.
prim_ping_exec(_TimeOut,'Ports.internalPort'(_WOut,-1,_,_WIn),_,
		'$io'('Just'(0)),E0,E) :-
	% ping to process port: must still be done....
	!, E0=E.
prim_ping_exec(TimeOut,'Ports.internalPort'(Host,SNr,PNr,_),_,
		'$io'(Result),E0,E) :-
	% ping external port
	string2Atom(Host,AHost),
	on_exception(_Exc,ping2SocketPort(AHost,SNr,PNr,TimeOut,Result),
                     Result='Nothing'), % ping fails in case of an error
	E0=E.

ping2SocketPort(Host,SNr,PNr,TimeOut,Result) :-
	socket('AF_INET',Socket),
	socket_connect(Socket,'AF_INET'(Host,SNr),PortStream),
	number_chars(PNr,PNrS),
	putChars(PortStream,[80|PNrS]), % 80 = 'P'
	put(PortStream,10),
	flush_output(PortStream),
	statistics(walltime,[ETime1,_]),
	ifTracePort((writeErr('TRACEPORTS: Ping to "'),
		     writeErr(SNr), writeErr('/'),
		     writeErr(PNr), writeErr('@'),
		     writeErr(Host), writeErr('"'),
		     nlErr)),
	TOSec is TimeOut//1000,
	TOMSec is (TimeOut mod 1000) * 1000, % in microseconds
	socket_select([],_,_,TOSec:TOMSec,[PortStream],SelStreams),
	(SelStreams = [PortStream]
	 -> readStreamLine(PortStream,_Line),
	    close(PortStream),
	    statistics(walltime,[ETime2,_]),
	    PingTime is ETime2-ETime1,
	    ifTracePort((writeErr('TRACEPORTS: Ping answer received in '),
		         writeErr(PingTime),
			 writeErr(' ms'), nlErr)),
	    Result = 'Just'(PingTime)
	  ; ifTracePort((writeErr('TRACEPORTS: No ping answer received'),
		         nlErr)),
	    Result = 'Nothing').


%%%% connect
  
?- block prim_connectPortAtSocket(?,?,?,?,?,-,?).
prim_connectPortAtSocket(RSNr,RPNr,RHost,_,
	               '$io'('Ports.internalPort'(Host,SNr,PNr,'()')),E0,E) :-
	deref(RSNr,SNr), deref(RPNr,PNr), derefAll(RHost,Host),
	!, E0=E.


send2Socket(Host,SNr,PNr,Msg) :-
	socket('AF_INET',Socket),
	socket_connect(Socket,'AF_INET'(Host,SNr),PortStream),
	number_chars(PNr,PNrS),
	putChars(PortStream,[78|PNrS]), % 78 = 'N'
	put(PortStream,10),
	show4sending(Msg,MsgString,RevVarIndexs),
	writeChars(PortStream,MsgString),
	put(PortStream,0), % send end-of-term marker
	flush_output(PortStream),
	ifTracePort((writeErr('TRACEPORTS: Sent to "'),
		     writeErr(SNr), writeErr('/'),
		     writeErr(PNr), writeErr('@'),
		     writeErr(Host), writeErr('": '),
		     writeChars(user_error,MsgString), nlErr
		     %writeErr(Msg), nlErr
		    )),
	rev(RevVarIndexs,VarIndexs),
	receive_extvar_bindings(PortStream,VarIndexs).

% read all variable bindings from the sockets instream:
receive_extvar_bindings(Str,[]) :-
	!,
	% all variable bindings received, so close the stream:
	ifTracePort((writeErr('TRACEPORTS: Closing send-connection'),
		     nlErr)),
	close(Str).
receive_extvar_bindings(Str,VIs) :-
	readStreamLine(Str,VarLine),
	VarLine = [86|IndexS],  % 86='V'
	number_chars(Index,IndexS),
	readStreamUntilZero(Str,MsgString),
	(deleteIndexVariable(Index,VIs,Var,VIs1)
	 -> true
	  ; writeErr('ERROR: Illegal binding for logical variable received:'),
	    nlErr,
	    VIs1 = VIs),
	(parse_received_message(nostream,MsgString,Msg)
         -> ifTracePort((writeErr('TRACEPORTS: Received binding for variable '),
                         writeErr(Index), writeErr(': '),
			 putChars(user_error,MsgString), nlErr
		         %writeErr(Msg), nlErr
			))
          ; writeErr('ERROR: Illegal message received (ignored): '),
            putChars(user_error,MsgString), nlErr),
	!,
	Var=Msg,  % unify received (ground!) term with sent variable
	receive_extvar_bindings(Str,VIs1).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% conversion of Curry terms into string representation:
% the final (result) argument is a list of variable/integer pairs
% of all variables occurring in the term and their uniquely assigned indices:
%

show4sending(T,S,VIs) :- numberVars(T,GT,[],VIs), show_term(GT,qualified,S,[]).


% get the assigned index for a variable:
getVariableIndex(V,[(W,Index)|_],Index) :- V==W, !.
getVariableIndex(V,[_|VIs],Index) :- getVariableIndex(V,VIs,Index).

% get the index assigned to a variable:
getIndexVariable(Index,[(V,Index)|_],V) :- !.
getIndexVariable(Index,[_|VIs],V) :- getIndexVariable(Index,VIs,V).

% get and delete the index assigned to a variable:
deleteIndexVariable(Index,[(V,Index)|VIs],V,VIs) :- !.
deleteIndexVariable(Index,[VI|VIs],V,[VI|VIs1]) :-
	deleteIndexVariable(Index,VIs,V,VIs1).


% conversion of string representations of Curry terms into Curry terms:
parse_received_message(Stream,String,Term) :-
	readTerm(String,qualified,T,GroundTerm), skipWhiteSpace(T,[]),
	extractVars(GroundTerm,Term,[],VIs),
	send_extvar_bindings(VIs,Dones,Stream),
	closeStreamAfterDones(Dones,Stream).

send_extvar_bindings([],[],_).
send_extvar_bindings([(V,I)|VIs],[Done|Dones],Stream) :-
	send_extvar_binding(Stream,V,I,Done),
	send_extvar_bindings(VIs,Dones,Stream).

% close the stream after all list elements are instantiated:
closeStreamAfterDones([],nostream) :- !.
closeStreamAfterDones([],Stream) :-
	ifTracePort((writeErr('TRACEPORTS: Closing receive-connection'),
		     nlErr)),
	close(Stream).
closeStreamAfterDones([Done|Dones],Stream) :-
	closeStreamAfterDones(Done,Dones,Stream).

?- block closeStreamAfterDones(-,?,?).
closeStreamAfterDones(_,Dones,Stream) :- closeStreamAfterDones(Dones,Stream).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% definitions for process communication:
%
% internal representation of the handle of the port to the process:
% 'Ports.internalPort'(WOut,-1,<unused>,WIn) where:
% WOut: a Prolog stream, the output of the process
% WIn : a Prolog stream, the input to the process

%%%%% open a connection to an external process:

?- block prim_openProcessPort(?,?,?,-,?).
prim_openProcessPort(RCmdString,_,'$io'('Ports.internalPort'(WOUT,-1,0,WIN)),
		     E0,E) :-
	derefAll(RCmdString,CmdString),
	string2Atom(CmdString,Cmd),
	exec(Cmd,[pipe(WIN),pipe(WOUT),std],_), E0=E.


?- block prim_openStdInOutPort(?,?,-,?).
prim_openStdInOutPort(_,'$io'('Ports.internalPort'(InStr,-1,0,OutStr)),E0,E) :-
	prolog_flag(user_input,InStr),
	prolog_flag(user_output,OutStr),
	E0=E.


% open a connection to a Unix socket as a port (similar to process ports):
%?- block 'prim_openSocketConnectPort'(-,?,?,?,?,?),
%         'prim_openSocketConnectPort'(?,-,?,?,?,?),
%         'prim_openSocketConnectPort'(?,?,?,?,-,?).
%'prim_openSocketConnectPort'(SNr,SHst,_,
%	  '$io'('Ports.internalPort'(SocketStream,-1,0,SocketStream)),E0,E) :-
%	string2Atom(SHst,Host), !,
%	socket('AF_INET',Socket),
%	socket_connect(Socket,'AF_INET'(Host,SNr),SocketStream), E0=E.


?- block prim_choiceSPEP(?,?,?,-,?), prim_choiceSPEP(?,-,?,?,?).
prim_choiceSPEP(RPort,Msgs,R,E0,E) :- deref(RPort,Port),
	prim_choiceSPEP_exec(Port,Msgs,R,E0,E).

prim_choiceSPEP_exec(SPPort,share(M),Result,E0,E) :-
        !,
	get_mutable(V,M),
	(V='$eval'(R) % external message stream has been already evaluated
	 -> E0=E, Result='Right'(R)
	  ; prim_choiceSPEP_exec(SPPort,V,CResult,E0,E),
	    (CResult='Left'(_)
  	     -> Result=CResult
	      ; CResult='Right'(S),
	        (compileWithSharing(variable) -> propagateShare(S,TResult) ; S=TResult),
	        Result='Right'(TResult),
	        update_mutable('$eval'(TResult),M))).
prim_choiceSPEP_exec(_,[M|Ms],'Right'([M|Ms]),E0,E) :-
	!, E0=E. % stream already evaluated
prim_choiceSPEP_exec('Ports.internalPort'(WOut,-1,_,_),[],Result,E0,E) :-
	% message stream is empty, look at stream port:
	readStreamLine(WOut,WOLine),
	map2(char_int,SPOutLine,WOLine),
	ifTracePort((writeErr('TRACEPORTS: SP_Get: '),
		     writeChars(user_error,SPOutLine),
		     nlErr)),
	Result='Left'(SPOutLine),
	E0=E,
	!.
prim_choiceSPEP_exec(SPort,'Ports.basicServerLoop'(Port),Result,E0,E) :-
        Port='Ports.internalPort'(_,_,PNr,_),
	checkIncomingPortStreams(PNr,PortStream),
	!,
	readStreamUntilZero(PortStream,MsgString),
	(compileWithSharing(function)
	 -> makeShare('Ports.basicServerLoop'(Port),BSL)
	  ; BSL = 'Ports.basicServerLoop'(Port)),
	(parse_received_message(PortStream,MsgString,Msg)
	 -> ifTracePort((writeErr('TRACEPORTS: Received message: '),
		         writeErr(Msg), nlErr)),
	    Result='Right'([Msg|BSL])
	  ; writeErr('ERROR: Illegal message received (ignored): '),
	    putChars(user_error,MsgString), nlErr,
	    prim_choiceSPEP_exec(SPort,BSL,Result,E0,E)),
	E0=E,
	!.
prim_choiceSPEP_exec('Ports.internalPort'(WOut,-1,_,WIn),
		     'Ports.basicServerLoop'(Port),Result,E0,E) :-
        Port='Ports.internalPort'(_,_,PNr,Socket),
	socket_select([Socket],PortStreams,Clients,off,[WOut],_SelStreams),
	(compileWithSharing(function)
	 -> makeShare('Ports.basicServerLoop'(Port),BSL)
	  ; BSL = 'Ports.basicServerLoop'(Port)),
	(PortStreams=[PortStream]
	 -> Clients=[Client],
	    ifTracePort((writeErr('TRACEPORTS: Connection to client: '),
		         writeErr(Client), nlErr)),
	    (readPortMessage(PNr,PortStream,MsgString)
	     -> (parse_received_message(PortStream,MsgString,Msg)
	         -> ifTracePort((writeErr('TRACEPORTS: Received message: '),
		                 writeErr(Msg), nlErr)),
		    Result='Right'([Msg|BSL])
	          ; writeErr('ERROR: Illegal message received (ignored): '),
		    putChars(user_error,MsgString), nlErr,
		    prim_choiceSPEP_exec('Ports.internalPort'(WOut,-1,_,WIn),
					 BSL,Result,E0,E))
	      ; % try reading next message:
		prim_choiceSPEP_exec('Ports.internalPort'(WOut,-1,_,WIn),
		                     BSL,Result,E0,E))
	  ; % SelStreams must be WOut:
            readStreamLine(WOut,WOLine),
	    map2(char_int,SPOutLine,WOLine),
	    ifTracePort((writeErr('TRACEPORTS: SP_Get: '),
		         writeChars(user_error,SPOutLine),
		         nlErr)),
	    Result='Left'(SPOutLine)),
	E0=E,
	!.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary operations for handling terms with variables:

% numberVars(T1,T2,[],VIs): replace each variable X in T1 by 'VAR'(i)
% where i is a unique index. VIs is a list of variable/integer pairs
% (e.g., it contains (X,i)):
numberVars(V,'VAR'(I),VIs,NVIs) :-
	var(V), !,
	(getVariableIndex(V,VIs,I) % has this variable already replaced?
	 -> NVIs=VIs
	  ; length(VIs,I), NVIs=[(V,I)|VIs] ).
numberVars(T,NT,VIs,NVIs) :-
	isShowableArg(T),
	T =.. [C|Args],
	numberVarsList(Args,NArgs,VIs,NVIs),
	NT =.. [C|NArgs].

numberVarsList([],[],VIs,VIs).
numberVarsList([Arg|Args],[NArg|NArgs],VIs,NVIs) :-
	numberVars(Arg,NArg,VIs,VIs1),
	numberVarsList(Args,NArgs,VIs1,NVIs).


% extractVars(T1,T2,[],VIs): replace each 'VAR'(i) in T1 by a new variable X
% where VIs is the list of variable/integer pairs (e.g., it contains (X,i))
extractVars(V,V,VIs,VIs) :- var(V), !.
extractVars('VAR'(I),Var,VIs,NVIs) :- !,
	(getIndexVariable(I,VIs,Var)
	 -> NVIs=VIs
	  ; NVIs=[(Var,I)|VIs]).
extractVars(T,NT,VIs,NVIs) :-
	T =.. [C|Args],
	extractVarsList(Args,NArgs,VIs,NVIs),
	NT =.. [C|NArgs].

extractVarsList([],[],VIs,VIs).
extractVarsList([Arg|Args],[NArg|NArgs],VIs,NVIs) :-
	extractVars(Arg,NArg,VIs,VIs1),
	extractVarsList(Args,NArgs,VIs1,NVIs).


% not yet implemented in Curry2Prolog:
?- block 'prim_openStdIn'(?,?,-,?).
'prim_openStdIn'(_,'$io'([]),E,E) :-
        raise_exception('Ports.openStdIn not implemented!').

?- block 'prim_after'(-,?,?,?).
'prim_after'(_,success,E,E) :-
        raise_exception('Ports.after not implemented!').
