------------------------------------------------------------------------------
--- A HtmlServer manages the event handlers used in
--- dynamic web pages on the server side.
--- Each event handler is encoded as a string and stored on the server side
--- with a unique key. Only this key is sent in the actual web page
--- to the client.
--- Event handlers are only valid for a particular time period
--- specified by <code>eventHandlerExpiration</code>, i.e., after that time
--- event handlers will be deleted.

module HtmlServer(storeHandlerStringWithCgiKey,retrieveHandlerStringWithCgiKey) where

import Ports
import Read(readNat)
import ReadShowTerm(readsQTerm)
import Time
import System

-- Stores a new event handler for a given cgi program and returns a new key.
storeHandlerStringWithCgiKey :: String -> String -> IO String
storeHandlerStringWithCgiKey evh cgikey = do
  ehSend CleanOldEventHandlers
  time <- getClockTime
  let expireDate = toDate (eventHandlerExpiration time)
  ehSend (StoreNewEventWithCgiKey key expireDate cgikey evh)
  return (show key ++ ' ':show expireDate)
 where
   key free

-- Retrieves a previously stored event handler for a cgi program.
-- Returns Nothing if the handler is no longer available, i.e., expired.
retrieveHandlerStringWithCgiKey :: String -> String -> IO (Maybe String)
retrieveHandlerStringWithCgiKey key cgikey = 
  let (nums,dates) = break (==' ') key
   in if null dates
      then return Nothing
      else let datel = readsQTerm dates
            in if null datel
               then return Nothing
               else ehSend (GetEventWithCgiKey (readNat nums) (fst (head datel)) cgikey info) >>
                    return info
 where info free

-- Specifies for a date a new date when the event handler expires.
eventHandlerExpiration :: ClockTime -> ClockTime
eventHandlerExpiration = addHours 1
--eventHandlerExpiration = addMinutes 1

-- Specifies for a date a new date when the next cleanup of event handlers
-- should be done.
nextCleanup :: ClockTime -> ClockTime
nextCleanup = addMinutes 5

---------------------------------------------------------------------------
-- The implementation of the event handler server via ports:

type Date = (Int,Int,Int,Int,Int,Int)

-- Messages for event handler server:
data EHSMsg = StoreNewEventWithCgiKey Int Date String String
            | GetEventWithCgiKey Int Date String (Maybe String)
            | CleanOldEventHandlers
            | ShowAllHandlers String
            | Stop

-- Start the event handler server via script:
startServer = do
  pakcshome <- getEnviron "PAKCSHOME"
  system (pakcshome++"/www/start")

-- Stop the event handler server (usually, not required):
stopServer = ehSend Stop

-- show all currently stored handlers:
showAllHandlers = do
  ehSend (ShowAllHandlers s)
  putStr s
 where s free


-- Open a symbolic port and start the event handler server:
serve = do
  msgs <- openNamedPort "ehserver"
  time <- getClockTime
  ehServer 0 (toDate time) [] msgs

-- The event handler server loop:
-- Argument 1: Current index for numbering new events
-- Argument 2: Next date when cleanup is necessary
-- Argument 3: The current event handlers (index,experiation date,cgikey,contents)
ehServer :: Int -> Date -> [(Int,Date,String,String)] -> [EHSMsg] -> IO ()
ehServer _ _ _ (Stop:_) = done
ehServer maxkey cleandate ehs (StoreNewEventWithCgiKey key time cgikey info : msgs) = do
  doSolve $ key =:= maxkey
  ehServer incMaxKey cleandate ((key,time,cgikey,info):ehs) msgs
 where
  incMaxKey = if maxkey>30000 then 0 else maxkey+1 -- to avoid overflows

ehServer maxkey cleandate ehs (GetEventWithCgiKey key time cgikey mbinfo : msgs)
  | mbinfo =:= searchEv ehs
  = ehServer maxkey cleandate ehs msgs
 where
   searchEv [] = Nothing
   searchEv ((n,t,c,i):es) = if key==n && time==t
                             then if c==cgikey then Just i else Nothing
                             else searchEv es

ehServer maxkey cleandate ehs (CleanOldEventHandlers : msgs) = do
  time <- getClockTime
  if compareDate (toDate time) cleandate == LT
   then ehServer maxkey cleandate ehs msgs
   else putStrLn (toDateString time ++": cleanup (current number of handlers: "++
                  show (length ehs) ++")") >>
        ehServer maxkey (toDate (nextCleanup time))
                 (filter (isNotExpired (toDate time)) ehs) msgs
 where
  isNotExpired time (_,etime,_,_) = compareDate time etime == LT

ehServer maxkey cleandate ehs (ShowAllHandlers s : msgs)
  | s =:= "Next cleanup: " ++ toDateString (dateToClockTime cleandate) ++ "\n" ++
          "Current event handlers:\n" ++ concatMap showWithoutCgiKey ehs
  = ehServer maxkey cleandate ehs msgs
 where
  showWithoutCgiKey (key,time,_,info) =
    "No. "++ show key ++" "++ show time ++": "++info++"\n"
  showWithCgiKey (key,time,cgi,info) = -- only for testing
    "No. "++ show key ++" / date: "++ show time ++" / key: \"" ++cgi++"\":\n  "++info++"\n"

-- send a message to local event handler server:
ehSend msg = do
  mbport <- connectPortRepeat 1000 done 1 "ehserver@localhost"
  maybe (startServer >> ehSend msg) (doSend msg) mbport

