------------------------------------------------------------------------------
--- Library with some useful extensions to the IO monad.
---
--- @author Michael Hanus
--- @version November 2004
------------------------------------------------------------------------------

module IOExts(execCmd,connectToSocket,choiceStreamOrMsgs,
              readCompleteFile,updateFile,
              exclusiveIO,setAssoc,getAssoc,
              IORef,newIORef,readIORef,writeIORef) where

import System
import IO(Handle)

--- An action that executes a command with a new default shell process.
--- The standard I/O streams of the new process (stdin,stdout,stderr)
--- are returned as handles so that they can be explicitly manipulated.
--- They should be closed with <code>IO.hClose</code> since they are not
--- closed automatically when the process terminates.
--- @param cmd - the shell command to be executed
--- @return the handles of the input/output/error streams of the new process
execCmd :: String -> IO (Handle,Handle,Handle)
execCmd cmd = seq (groundNormalForm cmd) (prim_execCmd cmd)

prim_execCmd :: String -> IO (Handle,Handle,Handle)
prim_execCmd external

--- Creates a new connection to a Unix socket.
--- @param host - the host name of the connection
--- @param snr - the socket number of the connection
--- @return the handle of the stream that is connected to the socket snr@host
---         and which is both readable and writable
connectToSocket :: String -> Int -> IO Handle
connectToSocket host snr = seq (groundNormalForm host)
                               (seq (ensureNotFree snr)
                                    (prim_connectToSocket host snr))

prim_connectToSocket :: String -> Int -> IO Handle
prim_connectToSocket external


--- This function implements a committed choice over the receiving
--- a line of characters from a stream and messages from an external port.
---
--- <EM>Note that the implementation of choiceStreamOrMsgs works only with
--- Sicstus-Prolog 3.8.5 or higher (due to a bug in previous versions
--- of Sicstus-Prolog).</EM>
---
--- @param handle - a handle of an input stream
--- @param msgs   - a stream of messages received via an external port (see Ports)
--- @return (Left handle) if the handle has some characters to be received
---         (Right msgs) if the stream msgs is instantiated
---                      with at least one new message at the head

choiceStreamOrMsgs :: Handle -> [msg] -> IO (Either Handle [msg])
choiceStreamOrMsgs external


--- An action that reads the complete contents of a file and returns it.
--- This action can be used instead of the (lazy) <code>readFile</code>
--- action if the contents of the file might be changed.
--- @param file - the name of the file
--- @return the complete contents of the file
readCompleteFile :: String -> IO String
readCompleteFile file = do
  s <- readFile file
  f s (return s)
 where
   f []     r = r
   f (_:cs) r = f cs r


--- An action that updates the contents of a file.
--- @param f - the function to transform the contents
--- @param file - the name of the file
updateFile :: (String -> String) -> String -> IO ()
updateFile f file = do
  s <- readCompleteFile file
  writeFile file (f s)


--- Forces the exclusive execution of an action via a lock file.
--- For instance, (exclusiveIO "myaction.lock" act) ensures that
--- the action "act" is not executed by two processes on the same
--- system at the same time.
--- @param lockfile - the name of a global lock file
--- @param action - the action to be exclusively executed
--- @return the result of the execution of the action
exclusiveIO :: String -> IO a -> IO a
exclusiveIO lockfile action = do
  system ("lockfile -1 "++lockfile)
  actionResult <- action
  system ("rm -f "++lockfile)
  return actionResult


--- Defines a global association between two strings.
--- Both arguments must be evaluable to ground terms before applying
--- this operation.
setAssoc :: String -> String -> IO ()
setAssoc key val = seq (groundNormalForm key)
                       (seq (groundNormalForm val) (prim_setAssoc key val))

prim_setAssoc :: String -> String -> IO ()
prim_setAssoc external


--- Gets the value associated to a string.
--- Nothing is returned if there does not exist an associated value.
getAssoc :: String -> IO (Maybe String)
getAssoc key = seq (groundNormalForm key) (prim_getAssoc key)

prim_getAssoc :: String -> IO (Maybe String)
prim_getAssoc external


--- Mutable variables containing values of some type.
--- The values are not evaluated when they are assigned to an IORef.
data IORef a = IORef a

--- Creates a new IORef with an initial values.
newIORef :: a -> IO (IORef a)
newIORef external

--- Reads the current value of an IORef.
readIORef :: IORef a -> IO a
readIORef external

--- Updates the value of an IORef.
writeIORef :: IORef a -> a -> IO ()
writeIORef external

