------------------------------------------------------------------------------
--- Library to access parts of the system environment.
---
--- @author Michael Hanus
--- @version June 2005
------------------------------------------------------------------------------

module System(Date(..),getDate,showDate,getCPUTime,getElapsedTime,
              getArgs,getEnviron,getHostname,getPID,getProgName,system,
              getLoadPath,getLoadPathForFile,findFileInLoadPath,
              sleep,evalTime,evalSpace) where

import FileGoodies

--- A date is presented in the following form:
--- (Date year month day hour minute second)
--- This type will not be supported in the future.
--- Use type Time.ClockTime!

data Date = Date Int Int Int Int Int Int

--- Returns the current date and time.
--- This function will not be supported in the future.
--- Use function Time.getClockTime!

getDate :: IO Date
getDate external

--- Transform a date into a readable form.
--- This function will not be supported in the future.
--- Use function Time.toDateString!

showDate :: Date -> String
showDate (Date y mo d h mi s) = digit2 mo ++"/"++ digit2 d ++"/"++ show y
                                ++ " " ++
                                digit2 h ++":"++ digit2 mi ++":"++ digit2 s
  where digit2 n = if n<10 then ['0',chr(ord '0' + n)]
                           else show n

--- Returns the current cpu time of the process in milliseconds.

getCPUTime :: IO Int
getCPUTime external
 
--- Returns the current elapsed time of the process in milliseconds.

getElapsedTime :: IO Int
getElapsedTime external
 
--- Returns the list of the program's command line arguments.
--- The program name is not included.

getArgs :: IO [String]
getArgs external
 
--- Returns the value of an environment variable.
--- The empty string is returned for undefined environment variables.

getEnviron :: String -> IO String
getEnviron evar = seq (groundNormalForm evar) (prim_getEnviron evar)
 
prim_getEnviron :: String -> IO String
prim_getEnviron external
 
--- Returns the hostname of the machine running this process.

getHostname :: IO String
getHostname external

--- Returns the process identifier of the current Curry process.

getPID :: IO Int
getPID external


--- Returns the name of the current program, i.e., the name of the
--- main module currently executed.

getProgName :: IO String
getProgName external


--- Executes a shell command and return with the exit code of the command.
--- An exit status of zero means successful execution.

system :: String -> IO Int
system cmd = seq (groundNormalForm cmd) (prim_system cmd)

prim_system :: String -> IO Int
prim_system external


--- Returns the current path (list of directory names) that is
--- used for loading modules.
getLoadPath :: IO [String]
getLoadPath = getLoadPathForFile "xxx"

--- Returns the current path (list of directory names) that is
--- used for loading modules w.r.t. a given main module file.
--- The directory prefix of the module file (or "." if there is
--- no such prefix) is the first element of the load path and the
--- remaining elements are determined by the environment variables
--- CURRYRPATH, PAKCSHOME, and PAKCSLIBPATH.

getLoadPathForFile :: String -> IO [String]
getLoadPathForFile modfile = do
  pakcshome <- getEnviron "PAKCSHOME"
  pakcspath <- getEnviron "PAKCSLIBPATH"
  currypath <- getEnviron "CURRYPATH"
  return $ [dirName modfile] ++
           (if null currypath then [] else splitPath currypath) ++
           (if null pakcspath
            then if null pakcshome then [] else [pakcshome++"/lib"]
            else splitPath pakcspath)

--- Adds a directory name to a file by looking up the current load path (CURRYPATH).
--- An error message is delivered if there is no such file.
findFileInLoadPath :: String -> IO String
findFileInLoadPath prog = do
  loadpath <- getLoadPathForFile prog
  getFileInPath (baseName prog) [""] loadpath


--- The evaluation of the action (sleep n) puts the Curry process
--- asleep for n seconds.

sleep :: Int -> IO ()
sleep n = seq (ensureNotFree n) (prim_sleep n)

prim_sleep :: Int -> IO ()
prim_sleep external

--- Evaluates the argument to normal form (and return the normal form)
--- and print the time needed for this evaluation on standard error.

evalTime :: a -> a
evalTime external

--- Evaluates the argument to normal form (and return the normal form)
--- and print the time and space needed for this evaluation on standard error.
--- During the evaluation, the garbage collector is turned off.

evalSpace :: a -> a
evalSpace external

